@@ -44,15 +44,11 @@ import warnings
 import zlib
 
 import pika
-from tornado import concurrent
 from pika import exceptions
-from tornado import gen
-from tornado import locks
+from tornado import concurrent, gen, locks
 import yaml
 
-from rejected import PYTHON26
-from rejected import data
-from rejected import log
+from rejected import data, log
 
 LOGGER = logging.getLogger(__name__)
 
@@ -90,60 +86,77 @@ class Consumer(object):
     """Base consumer class that defines the contract between rejected and
     consumer applications.
 
-    In any of the consumer base classes, if the ``MESSAGE_TYPE`` attribute is
-    set, the ``type`` property of incoming messages will be validated against
-    when a message is received, checking for string equality against the
-    ``MESSAGE_TYPE`` attribute. If they are not matched, the consumer will not
+    In any of the consumer base classes, if the ``message_type`` is specified
+    in the configuration (or set with the ``MESSAGE_TYPE`` attribute), the
+    ``type`` property of incoming messages will be validated against when a
+    message is received. If there is no match, the consumer will not
     process the message and will drop the message without an exception if the
-    ``DROP_INVALID_MESSAGES`` attribute is set to ``True``. If it is ``False``,
-    a :py:class:`MessageException` is raised.
+    ``drop_invalid_messages`` setting is set to ``True`` in the configuration
+    (or if the ``DROP_INVALID_MESSAGES`` attribute is set to ``True``).
+    If it is ``False``, a :exc:`~rejected.consumer.MessageException` is raised.
 
-    If a consumer raises a :py:class:`ProcessingException`, the message that
-    was being processed will be republished to the exchange specified by the
-    ``ERROR_EXCHANGE`` attribute of the consumer's class using the routing key
-    that was last used for the message. The original message body and
-    properties will be used and an additional header
+    If a consumer raises a `~rejected.consumer.ProcessingException`, the
+    message that was being processed will be republished to the exchange
+    specified by the ``error`` exchange configuration value or the
+    ``ERROR_EXCHANGE`` attribute of the consumer's class. The message will be
+    published using the routing key that was last used for the message. The
+    original message body and properties will be used and an additional header
     ``X-Processing-Exceptions`` will be added that will contain the number of
-    times the message has had a ``ProcessingException`` raised for it. In
-    combination with a queue that has ``x-message-ttl`` set and
-    ``x-dead-letter-exchange`` that points to the original exchange for the
+    times the message has had a :exc:`~rejected.consumer.ProcessingException`
+    raised for it. In combination with a queue that has ``x-message-ttl`` set
+    and ``x-dead-letter-exchange`` that points to the original exchange for the
     queue the consumer is consuming off of, you can implement a delayed retry
     cycle for messages that are failing to process due to external resource or
     service issues.
 
-    If ``ERROR_MAX_RETRY`` is set on the class, the headers for each method
+    If ``error_max_retry`` is specified in the configuration or
+    ``ERROR_MAX_RETRY`` is set on the class, the headers for each method
     will be inspected and if the value of ``X-Processing-Exceptions`` is
-    greater than or equal to the ``ERROR_MAX_RETRY`` value, the message will
+    greater than or equal to the specified value, the message will
     be dropped.
 
+    :param dict settings: The configuration from rejected
+    :param rejected.process.Process process: The controlling process
+    :param bool drop_invalid_messages: Drop a message if its type property
+        doesn't match the specified message type.
+    :param str|list message_type: Used to validate the message type of a
+        message before processing. This attribute can be set to a string
+        that is matched against the AMQP message type or a list of
+        acceptable message types.
+    :param error_exchange: The exchange to publish a message raising a
+        :exc:`~rejected.consumer.ProcessingException` to
+    :type error_exchange: str
+    :param int error_max_retry: The number of
+        :exc:`~rejected.consumer.ProcessingException`s raised on a message
+        before a message is dropped. If not specified, messages will never be
+        dropped.
+
     """
     DROP_INVALID_MESSAGES = False
-    """Drop a message if its type property doesn't match ``MESSAGE_TYPE``"""
-
     MESSAGE_TYPE = None
-    """Used to validate the message type of a message before processing"""
-
     ERROR_EXCHANGE = 'errors'
-    """The exchange to publish ProcessingErrors to"""
-
     ERROR_MAX_RETRY = None
-    """The number of ``ProcessingErrors` before a message is dropped"""
 
-    def __init__(self, settings, process):
+    def __init__(self, settings, process,
+                 drop_invalid_messages=DROP_INVALID_MESSAGES,
+                 message_type=MESSAGE_TYPE,
+                 error_exchange=ERROR_EXCHANGE,
+                 error_max_retry=ERROR_MAX_RETRY):
         """Creates a new instance of a Consumer class. To perform
         initialization tasks, extend Consumer.initialize
 
-        :param dict settings: The configuration from rejected
-        :param rejected.process.Process: The controlling process
-
         """
         self._channel = None
+        self._drop_invalid = drop_invalid_messages
+        self._error_exchange = error_exchange
+        self._error_max_retry = error_max_retry
         self._finished = False
         self._message = None
+        self._message_type = message_type
+        self._measurement = None
         self._message_body = None
         self._process = process
         self._settings = settings
-        self._statsd = None
         self._yield_condition = locks.Condition()
 
         # Create a logger that attaches correlation ID to the record
@@ -154,12 +167,6 @@ class Consumer(object):
         # Set a Sentry context for the consumer
         self.set_sentry_context('consumer', self.name)
 
-        try:
-            self.set_sentry_context('version',
-                                    sys.modules[self.__module__].__version__)
-        except (NameError, AttributeError):
-            pass
-
         # Run any child object specified initialization
         self.initialize()
 
@@ -235,29 +242,83 @@ class Consumer(object):
             self.logger.debug('Setting sentry context for %s to %s', tag, value)
             self.sentry_client.tags_context({tag: value})
 
-    def statsd_add_timing(self, key, duration):
-        """Add a timing to statsd
+    def stats_add_timing(self, key, duration):
+        """Add a timing to the per-message measurements
 
         :param str key: The key to add the timing to
         :param int|float duration: The timing value
 
         """
-        if self._statsd:
-            self._statsd.add_timing(key, duration)
+        if not self._measurement:
+            LOGGER.warning('stats_add_timing invoked outside execution')
+            return
+        self._measurement.set_value(key, duration)
 
-    def statsd_incr(self, key, value=1):
-        """Increment the specified key in statsd if statsd is enabled.
+    def statsd_add_timing(self, key, duration):
+        """Add a timing to the per-message measurements
+
+        :param str key: The key to add the timing to
+        :param int|float duration: The timing value
+
+        .. deprecated:: 3.13.0
+
+        """
+        warnings.warn('Deprecated, use Consumer.stats_add_timing',
+                      DeprecationWarning)
+        self.stats_add_timing(key, duration)
+
+    def stats_incr(self, key, value=1):
+        """Increment the specified key in the per-message measurements
 
         :param str key: The key to increment
         :param int value: The value to increment the key by
 
         """
-        if self._statsd:
-            self._statsd.incr(key, value)
+        if not self._measurement:
+            LOGGER.warning('stats_incr invoked outside execution')
+            return
+        self._measurement.incr(key, value)
+
+    def statsd_incr(self, key, value=1):
+        """Increment the specified key in the per-message measurements
+
+        :param str key: The key to increment
+        :param int value: The value to increment the key by
+
+        .. deprecated:: 3.13.0
+
+        """
+        warnings.warn('Deprecated, use Consumer.stats_incr',
+                      DeprecationWarning)
+        self.stats_incr(key, value)
+
+    def stats_set_tag(self, key, value=1):
+        """Set the specified tag/value in the per-message measurements
+
+        :param str key: The key to increment
+        :param int value: The value to increment the key by
+
+        """
+        if not self._measurement:
+            LOGGER.warning('stats_set_tag invoked outside execution')
+            return
+        self._measurement.set_tag(key, value)
+
+    def stats_set_value(self, key, value=1):
+        """Set the specified key/value in the per-message measurements
+
+        :param str key: The key to increment
+        :param int value: The value to increment the key by
+
+        """
+        if not self._measurement:
+            LOGGER.warning('stats_set_value invoked outside execution')
+            return
+        self._measurement.set_value(key, value)
 
     @contextlib.contextmanager
-    def statsd_track_duration(self, key):
-        """Time around a context and emit a statsd metric.
+    def stats_track_duration(self, key):
+        """Time around a context and add to the the per-message measurements
 
         :param str key: The key for the timing to track
 
@@ -266,8 +327,20 @@ class Consumer(object):
         try:
             yield
         finally:
-            finish_time = max(start_time, time.time())
-            self.statsd_add_timing(key, finish_time - start_time)
+            self.stats_add_timing(
+                key, max(start_time, time.time()) - start_time)
+
+    def statsd_track_duration(self, key):
+        """Time around a context and add to the the per-message measurements
+
+        :param str key: The key for the timing to track
+
+        .. deprecated:: 3.13.0
+
+        """
+        warnings.warn('Deprecated, use Consumer.stats_track_duration',
+                      DeprecationWarning)
+        return self.stats_track_duration(key)
 
     def unset_sentry_context(self, tag):
         """Remove a context tag from sentry
@@ -545,34 +618,43 @@ class Consumer(object):
         self._message_body = None
 
     @gen.coroutine
-    def _execute(self, message_in):
+    def _execute(self, message_in, measurement):
         """Process the message from RabbitMQ. To implement logic for processing
         a message, extend Consumer._process, not this method.
 
         :param rejected.Consumer.Message message_in: The message to process
+        :param measurement: For collecting per-message instrumentation
+        :type measurement: rejected.data.Measurement
         :rtype: bool
 
         """
         LOGGER.debug('Received: %r', message_in)
         self._clear()
         self._message = message_in
+        self._measurement = measurement
 
         if self.message_type:
             self.set_sentry_context('type', self.message_type)
 
-        # Validate the message type if the child sets _MESSAGE_TYPE
-        if self.MESSAGE_TYPE and self.MESSAGE_TYPE != self.message_type:
-            self.logger.warning('Received a non-supported message type: %s',
-                                self.message_type)
-            # Should the message be dropped or returned to the broker?
-            if self.DROP_INVALID_MESSAGES:
-                raise gen.Return(data.MESSAGE_DROP)
-            raise gen.Return(data.MESSAGE_EXCEPTION)
+        # Validate the message type if the child sets MESSAGE_TYPE
+        if self._message_type:
+            if isinstance(self._message_type, (tuple, list, set)):
+                message_supported = self.message_type in self._message_type
+            else:
+                message_supported = self.message_type == self._message_type
+
+            if not message_supported:
+                self.logger.warning('Received unsupported message type: %s',
+                                    self.message_type)
+                # Should the message be dropped or returned to the broker?
+                if self._drop_invalid:
+                    raise gen.Return(data.MESSAGE_DROP)
+                raise gen.Return(data.MESSAGE_EXCEPTION)
 
         # Check the number of ProcessingErrors and possibly drop the message
-        if (self.ERROR_MAX_RETRY and
+        if (self._error_max_retry and
                 _PROCESSING_EXCEPTIONS in self.headers):
-            if self.headers[_PROCESSING_EXCEPTIONS] >= self.ERROR_MAX_RETRY:
+            if self.headers[_PROCESSING_EXCEPTIONS] >= self._error_max_retry:
                 self.logger.warning('Dropping message with %i deaths due to '
                                     'ERROR_MAX_RETRY',
                                     self.headers[_PROCESSING_EXCEPTIONS])
@@ -599,13 +681,11 @@ class Consumer(object):
         except exceptions.ChannelClosed as error:
             self.logger.critical('Channel closed while processing %s: %s',
                                  message_in.delivery_tag, error)
-            self._process.reconnect()
             raise gen.Return(None)
 
         except exceptions.ConnectionClosed as error:
             self.logger.critical('Connection closed while processing %s: %s',
                                  message_in.delivery_tag, error)
-            self._process.reconnect()
             raise gen.Return(None)
 
         except ConsumerException as error:
@@ -654,7 +734,7 @@ class Consumer(object):
             except TypeError:
                 properties['headers'][_PROCESSING_EXCEPTIONS] = 1
 
-        self._channel.basic_publish(self.ERROR_EXCHANGE,
+        self._channel.basic_publish(self._error_exchange,
                                     self._message.routing_key,
                                     self._message.body,
                                     pika.BasicProperties(**properties))
@@ -668,14 +748,6 @@ class Consumer(object):
         """
         self._channel = channel
 
-    def _set_statsd(self, statsd):
-        """Assign a `StatsdClient` instance to the class.
-
-        :param pika.statsd.StatsdClient statsd: The StatsdClient instance
-
-        """
-        self._statsd = statsd
-
     def log_exception(self, msg_format, *args, **kwargs):
         """Customize the logging of uncaught exceptions.
 
@@ -696,12 +768,8 @@ class Consumer(object):
         if all(exc_info):
             exc_type, exc_value, tb = exc_info
             exc_name = exc_type.__name__
-            if PYTHON26:
-                self.logger.exception('Processor handled %s: %s', exc_name,
-                                      exc_value)
-            else:
-                self.logger.exception('Processor handled %s: %s', exc_name,
-                                      exc_value, exc_info=exc_info)
+            self.logger.exception('Processor handled %s: %s', exc_name,
+                                  exc_value)
 
         if kwargs.get('send_to_sentry', True):
             self._process.send_exception_to_sentry(exc_info)
@@ -733,6 +801,20 @@ class PublishingConsumer(Consumer):
     ``DROP_INVALID_MESSAGES`` attribute is set to ``True``. If it is ``False``,
     a :py:class:`ConsumerException` is raised.
 
+    :param dict settings: The configuration from rejected
+    :param rejected.process.Process process: The controlling process
+    :param bool drop_invalid_messages: Drop a message if its type property
+        doesn't match the specified message type.
+    :param str|list message_type: Used to validate the message type of a
+        message before processing. This attribute can be set to a string
+        that is matched against the AMQP message type or a list of
+        acceptable message types.
+    :param error_exchange: The exchange to publish `ProcessingException` to
+    :type error_exchange: str
+    :param int error_max_retry: The number of `ProcessingException`s
+        raised on a message before a message is dropped. If not specified,
+        messages will never be dropped.
+
     """
 
     def initialize(self):
@@ -754,8 +836,8 @@ class PublishingConsumer(Consumer):
 
         # Publish the message
         self.logger.debug('Publishing message to %s:%s', exchange, routing_key)
-        with self.statsd_track_duration('publish.{}.{}'.format(exchange,
-                                                               routing_key)):
+        with self._measurement.track_duration(
+                'publish.{}.{}'.format(exchange, routing_key)):
             self._channel.basic_publish(exchange=exchange,
                                         routing_key=routing_key,
                                         properties=msg_props,
@@ -858,6 +940,20 @@ class SmartConsumer(Consumer):
     ``DROP_INVALID_MESSAGES`` attribute is set to ``True``. If it is ``False``,
     a :py:class:`ConsumerException` is raised.
 
+    :param dict settings: The configuration from rejected
+    :param rejected.process.Process process: The controlling process
+    :param bool drop_invalid_messages: Drop a message if its type property
+        doesn't match the specified message type.
+    :param str|list message_type: Used to validate the message type of a
+        message before processing. This attribute can be set to a string
+        that is matched against the AMQP message type or a list of
+        acceptable message types.
+    :param error_exchange: The exchange to publish `ProcessingException` to
+    :type error_exchange: str
+    :param int error_max_retry: The number of `ProcessingException`s
+        raised on a message before a message is dropped. If not specified,
+        messages will never be dropped.
+
     """
 
     @property
@@ -913,7 +1009,7 @@ class SmartConsumer(Consumer):
     def _decode_bz2(value):
         """Return a bz2 decompressed value
 
-        :param str value: Compressed value
+        :param bytes value: Compressed value
         :rtype: str
 
         """
@@ -923,7 +1019,7 @@ class SmartConsumer(Consumer):
     def _decode_gzip(value):
         """Return a zlib decompressed value
 
-        :param str value: Compressed value
+        :param bytes value: Compressed value
         :rtype: str
 
         """
@@ -940,6 +1036,8 @@ class SmartConsumer(Consumer):
         """
         if not bs4:
             raise ConsumerException('BeautifulSoup4 is not enabled')
+        if isinstance(value, bytes):
+            value = value.decode('utf-8')
         return bs4.BeautifulSoup(value)
 
     @staticmethod
@@ -951,6 +1049,8 @@ class SmartConsumer(Consumer):
         :rtype: csv.DictReader
 
         """
+        if isinstance(value, bytes):
+            value = value.decode('utf-8')
         csv_buffer = io.StringIO(value)
         dialect = csv.Sniffer().sniff(csv_buffer.read(1024))
         csv_buffer.seek(0)
@@ -964,11 +1064,12 @@ class SmartConsumer(Consumer):
         :rtype: object
 
         """
+        if isinstance(value, bytes):
+            value = value.decode('utf-8')
         try:
             return json.loads(value, encoding='utf-8')
         except ValueError as error:
-            self.logger.error('Could not decode message body: %s', error,
-                              exc_info=sys.exc_info())
+            self.logger.exception('Could not decode message body: %s', error)
             raise MessageException(error)
 
     def _load_msgpack_value(self, value):
@@ -982,8 +1083,7 @@ class SmartConsumer(Consumer):
         try:
             return umsgpack.unpackb(value)
         except ValueError as error:
-            self.logger.error('Could not decode message body: %s', error,
-                              exc_info=sys.exc_info())
+            self.logger.exception('Could not decode message body: %s', error)
             raise MessageException(error)
 
     @staticmethod
@@ -991,7 +1091,7 @@ class SmartConsumer(Consumer):
         """Deserialize a pickle string returning the native Python data type
         for the value.
 
-        :param str value: The pickle string
+        :param bytes value: The pickle string
         :rtype: object
 
         """
@@ -1002,7 +1102,7 @@ class SmartConsumer(Consumer):
         """Deserialize a plist string returning the native Python data type
         for the value.
 
-        :param str value: The pickle string
+        :param bytes value: The pickle string
         :rtype: dict
 
         """
@@ -1026,8 +1126,23 @@ class SmartConsumer(Consumer):
 
 
 class SmartPublishingConsumer(SmartConsumer, PublishingConsumer):
-    """PublishingConsumer with serialization built in"""
+    """PublishingConsumer with serialization built in
 
+    :param dict settings: The configuration from rejected
+    :param rejected.process.Process process: The controlling process
+    :param bool drop_invalid_messages: Drop a message if its type property
+        doesn't match the specified message type.
+    :param str|list message_type: Used to validate the message type of a
+        message before processing. This attribute can be set to a string
+        that is matched against the AMQP message type or a list of
+        acceptable message types.
+    :param error_exchange: The exchange to publish `ProcessingException` to
+    :type error_exchange: str
+    :param int error_max_retry: The number of `ProcessingException`s
+        raised on a message before a message is dropped. If not specified,
+        messages will never be dropped.
+
+    """
     def publish_message(self, exchange, routing_key, properties, body,
                         no_serialization=False,
                         no_encoding=False):
@@ -1172,17 +1287,17 @@ class SmartPublishingConsumer(SmartConsumer, PublishingConsumer):
         """Serialize a value into JSON
 
         :param str|dict|list: The value to serialize as JSON
-        :rtype: str
+        :rtype: bytes
 
         """
-        return json.dumps(value, ensure_ascii=False)
+        return json.dumps(value, ensure_ascii=True).encode('utf-8')
 
     @staticmethod
     def _dump_msgpack_value(value):
         """Serialize a value into MessagePack
 
         :param str|dict|list: The value to serialize as msgpack
-        :rtype: str
+        :rtype: bytes
 
         """
         return umsgpack.packb(value)
@@ -1192,7 +1307,7 @@ class SmartPublishingConsumer(SmartConsumer, PublishingConsumer):
         """Serialize a value into the pickle format
 
         :param any value: The object to pickle
-        :rtype: str
+        :rtype: bytes
 
         """
         return pickle.dumps(value)
@@ -1202,13 +1317,13 @@ class SmartPublishingConsumer(SmartConsumer, PublishingConsumer):
         """Create a plist value from a dictionary
 
         :param dict value: The value to make the plist from
-        :rtype: dict
+        :rtype: bytes
 
         """
         if hasattr(plistlib, 'dumps'):
             return plistlib.dumps(value)
         try:
-            return plistlib.writePlistToString(value)
+            return plistlib.writePlistToString(value).encode('utf-8')
         except AttributeError:
             return plistlib.writePlistToBytes(value)
 
@@ -1227,9 +1342,11 @@ class SmartPublishingConsumer(SmartConsumer, PublishingConsumer):
         """Return a bzip2 compressed value
 
         :param str value: Uncompressed value
-        :rtype: str
+        :rtype: bytes
 
         """
+        if not isinstance(value, bytes):
+            value = value.encode('utf-8')
         return bz2.compress(value)
 
     @staticmethod
@@ -1237,9 +1354,11 @@ class SmartPublishingConsumer(SmartConsumer, PublishingConsumer):
         """Return zlib compressed value
 
         :param str value: Uncompressed value
-        :rtype: str
+        :rtype: bytes
 
         """
+        if not isinstance(value, bytes):
+            value = value.encode('utf-8')
         return zlib.compress(value)
 
 
