@@ -3,6 +3,7 @@ Consumer process management. Imports consumer code, manages RabbitMQ
 connection state and collects stats about the consuming process.
 
 """
+import collections
 import importlib
 import logging
 import math
@@ -14,14 +15,16 @@ try:
     import cProfile as profile
 except ImportError:
     import profile
-import re
 import signal
 import sys
 import time
+import warnings
 
-from tornado import gen
-from tornado import ioloop
-from tornado import locks
+try:
+    import sprockets_influxdb as influxdb
+except ImportError:
+    influxdb = None
+from tornado import gen, ioloop, locks
 import pika
 try:
     import raven
@@ -29,16 +32,10 @@ except ImportError:
     raven = None
 from pika import spec
 
-from rejected import __version__
-from rejected import PYTHON26
-from rejected import data
-from rejected import state
-from rejected import stats
+from rejected import __version__, data, state, statsd
 
 LOGGER = logging.getLogger(__name__)
 
-URI_RE = re.compile(r'^[\w\+\-]+://.*:(\w+)@.*')
-
 
 def import_consumer(value):
     """Pass in a string in the format of foo.Bar, foo.bar.Baz, foo.bar.baz.Qux
@@ -76,16 +73,17 @@ class Process(multiprocessing.Process, state.State):
     # Counter constants
     ACKED = 'acked'
     CLOSED_ON_COMPLETE = 'closed_on_complete'
+    DROPPED = 'dropped'
     ERROR = 'failed'
     FAILURES = 'failures_until_stop'
+    NACKED = 'nacked'
     PROCESSED = 'processed'
-    REDELIVERED = 'redelivered_messages'
+    REQUEUED = 'requeued'
+    REDELIVERED = 'redelivered'
     TIME_SPENT = 'processing_time'
     TIME_WAITED = 'idle_time'
 
     MESSAGE_AGE = 'message_age'
-    MESSAGE_DROPPED = 'message_dropped'
-    MESSAGE_REQUEUED = 'message_requeued'
 
     CONSUMER_EXCEPTION = 'consumer_exception'
     MESSAGE_EXCEPTION = 'message_exception'
@@ -122,21 +120,26 @@ class Process(multiprocessing.Process, state.State):
         self.consumer = None
         self.consumer_lock = None
         self.consumer_name = None
+        self.consumer_version = None
+        self.counters = collections.Counter()
         self.delivery_time = None
+        self.influxdb = None
         self.ioloop = None
         self.last_failure = 0
         self.last_stats_time = None
         self.logging_config = dict()
+        self.measurement = None
         self.message_connection_id = None
         self.max_error_count = self.MAX_ERROR_COUNT
         self.max_frame_size = spec.FRAME_MAX_SIZE
         self.queue_name = None
         self.prepend_path = None
+        self.previous = None
         self.sentry_client = None
         self.state = self.STATE_INITIALIZING
         self.state_start = time.time()
-        self.stats = None
         self.stats_queue = None
+        self.statsd = None
 
         # Override ACTIVE with PROCESSING
         self.STATES[0x04] = 'Processing'
@@ -149,11 +152,12 @@ class Process(multiprocessing.Process, state.State):
         """
         if not self.can_respond:
             LOGGER.warning('Can not ack message, disconnected from RabbitMQ')
-            self.stats.incr(self.CLOSED_ON_COMPLETE)
+            self.counters[self.CLOSED_ON_COMPLETE] += 1
             return
         LOGGER.debug('Acking %s', delivery_tag)
         self.channel.basic_ack(delivery_tag=delivery_tag)
-        self.stats.incr(self.ACKED)
+        self.counters[self.ACKED] += 1
+        self.measurement.set_tag(self.ACKED, True)
 
     def add_on_channel_close_callback(self):
         """This method tells pika to call the on_channel_closed method if
@@ -274,33 +278,33 @@ class Process(multiprocessing.Process, state.State):
         try:
             consumer_, version = import_consumer(cfg['consumer'])
         except ImportError as error:
-            if PYTHON26:
-                LOGGER.exception('Error importing the consumer %s: %s',
-                                 cfg['consumer'], error)
-            else:
-                LOGGER.exception('Error importing the consumer %s: %s',
-                                 cfg['consumer'], error,
-                                 exc_info=sys.exc_info())
+            LOGGER.exception('Error importing the consumer %s: %s',
+                             cfg['consumer'], error)
             return
 
         if version:
             LOGGER.info('Creating consumer %s v%s', cfg['consumer'], version)
+            self.consumer_version = version
         else:
             LOGGER.info('Creating consumer %s', cfg['consumer'])
 
         settings = cfg.get('config', dict())
         settings['_import_module'] = '.'.join(cfg['consumer'].split('.')[0:-1])
 
+        kwargs = {
+            'settings': settings,
+            'process': self,
+            'drop_invalid_messages': cfg.get('drop_invalid_messages'),
+            'message_type': cfg.get('message_type'),
+            'error_exchange': cfg.get('error_exchange'),
+            'error_max_retry': cfg.get('error_max_retry')
+        }
+
         try:
-            return consumer_(settings=settings, process=self)
+            return consumer_(**kwargs)
         except Exception as error:
-            if PYTHON26:
-                LOGGER.exception('Error creating the consumer "%s": %s',
-                                 cfg['consumer'], error)
-            else:
-                LOGGER.exception('Error creating the consumer "%s": %s',
-                                 cfg['consumer'], error,
-                                 exc_info=sys.exc_info())
+            LOGGER.exception('Error creating the consumer "%s": %s',
+                             cfg['consumer'], error)
 
     def get_module_data(self):
         modules = {}
@@ -340,24 +344,26 @@ class Process(multiprocessing.Process, state.State):
                 self.delivery_time = start_time = time.time()
                 self.active_message = message
 
-                # Track the message age
+                self.measurement = data.Measurement()
+
+                if message.method.redelivered:
+                    self.counters[self.REDELIVERED] += 1
+                    self.measurement.set_tag(self.REDELIVERED, True)
+
                 if message.properties.timestamp:
-                    self.stats.add_timing(self.MESSAGE_AGE,
-                                          start_time -
-                                          message.properties.timestamp)
+                    self.measurement.set_value(
+                        self.MESSAGE_AGE,
+                        max(message.properties.timestamp, start_time) -
+                        message.properties.timestamp)
 
                 self.start_message_processing()
                 try:
-                    result = yield self.consumer._execute(message)
+                    result = yield self.consumer._execute(message,
+                                                          self.measurement)
                 except Exception as error:
-                    if PYTHON26:
-                        LOGGER.exception('Unhandled exception from consumer in '
-                                         'process. This should not happen. %s',
-                                         error)
-                    else:
-                        LOGGER.exception('Unhandled exception from consumer in '
-                                         'process. This should not happen. %s',
-                                         error, exc_info=sys.exc_info())
+                    LOGGER.exception('Unhandled exception from consumer in '
+                                     'process. This should not happen. %s',
+                                     error)
                     result = data.MESSAGE_REQUEUE
 
                 LOGGER.debug('Finished processing message: %r', result)
@@ -381,6 +387,16 @@ class Process(multiprocessing.Process, state.State):
         """
         return self.state in [self.STATE_PROCESSING, self.STATE_STOP_REQUESTED]
 
+    def maybe_submit_measurement(self):
+        """Check for configured instrumentation backends and if found, submit
+        the message measurement info.
+
+        """
+        if self.statsd:
+            self.submit_statsd_measurements()
+        if self.influxdb:
+            self.submit_influxdb_measurement()
+
     def on_channel_closed(self, _channel, reply_code, reply_text):
         """Invoked by pika when RabbitMQ unexpectedly closes the channel.
         Channels are usually closed if you attempt to do something that
@@ -450,63 +466,63 @@ class Process(multiprocessing.Process, state.State):
         :param str body: The message body
 
         """
-        if method.redelivered:
-            self.stats.incr(self.REDELIVERED)
-        try:
-            self.invoke_consumer(data.Message(channel, method, properties,
-                                              body))
-        except AttributeError:
-            pass
+        self.invoke_consumer(data.Message(channel, method, properties, body))
 
     def on_processed(self, message, result, start_time):
-        self.stats.add_timing(self.TIME_SPENT, time.time() - start_time)
+        """Invoked after a message is processed by the consumer and
+        implements the logic for how to deal with a message based upon
+        the result.
+
+        :param rejected.data.Message message: The message that was processed
+        :param int result: The result of the processing of the message
+        :param float start_time: When the message was received
+
+        """
+        duration = max(start_time, time.time()) - start_time
+        self.counters[self.TIME_SPENT] += duration
+        self.measurement.set_value(self.TIME_SPENT, duration)
 
         if result == data.MESSAGE_DROP:
             LOGGER.debug('Rejecting message due to drop return from consumer')
             self.reject(message.delivery_tag, False)
-            self.stats.incr(self.MESSAGE_DROPPED)
-            return
+            self.counters[self.DROPPED] += 1
 
         elif result == data.MESSAGE_EXCEPTION:
-            LOGGER.debug('Rejecteting due to MessageException')
+            LOGGER.debug('Rejecting message due to MessageException')
             self.reject(message.delivery_tag, False)
-            self.stats.incr(self.MESSAGE_EXCEPTION)
-            return
+            self.counters[self.MESSAGE_EXCEPTION] += 1
+            self.measurement.set_tag('exception', 'MessageException')
 
         elif result == data.PROCESSING_EXCEPTION:
             LOGGER.debug('Rejecting message due to ProcessingException')
             self.reject(message.delivery_tag, False)
-            self.stats.incr(self.PROCESSING_EXCEPTION)
-            return
+            self.counters[self.PROCESSING_EXCEPTION] += 1
+            self.measurement.set_tag('exception', 'ProcessingException')
 
         elif result == data.CONSUMER_EXCEPTION:
             LOGGER.debug('Re-queueing message due to ConsumerException')
             self.reject(message.delivery_tag, True)
             self.on_processing_error()
-            self.stats.incr(self.CONSUMER_EXCEPTION)
-            return
+            self.counters[self.CONSUMER_EXCEPTION] += 1
+            self.measurement.set_tag('exception', 'ConsumerException')
 
         elif result == data.UNHANDLED_EXCEPTION:
             LOGGER.debug('Re-queueing message due to UnhandledException')
             self.reject(message.delivery_tag, True)
             self.on_processing_error()
-            self.stats.incr(self.UNHANDLED_EXCEPTION)
-            return
+            self.counters[self.UNHANDLED_EXCEPTION] += 1
+            self.measurement.set_tag('exception', 'UnhandledException')
 
         elif result == data.MESSAGE_REQUEUE:
             LOGGER.debug('Re-queueing message due Consumer request')
             self.reject(message.delivery_tag, True)
-            self.stats.incr(self.MESSAGE_REQUEUED)
-            return
+            self.counters[self.REQUEUED] += 1
 
-        # Ack if the msg wasn't rejected by MessageException and self.ack = True
         elif result == data.MESSAGE_ACK and self.ack:
             self.ack_message(message.delivery_tag)
-            self.stats.incr(self.PROCESSED)
-
-        if self.is_waiting_to_shutdown:
-            return self.on_ready_to_stop()
 
+        self.counters[self.PROCESSED] += 1
+        self.maybe_submit_measurement()
         self.reset_state()
 
     def on_processing_error(self):
@@ -519,11 +535,11 @@ class Process(multiprocessing.Process, state.State):
             LOGGER.info('Resetting failure window, %i seconds since last',
                         duration)
             self.reset_error_counter()
-        self.stats.incr(self.ERROR)
+        self.counters[self.ERROR] += 1
         self.last_failure = time.time()
         if self.too_many_errors:
             LOGGER.critical('Error threshold exceeded (%i), shutting down',
-                            self.stats[self.ERROR])
+                            self.counters[self.ERROR])
             self.cancel_consumer_with_rabbitmq()
             self.close_connection()
             self.on_ready_to_stop()
@@ -576,8 +592,7 @@ class Process(multiprocessing.Process, state.State):
         :param frame _unused_frame: The python frame the signal was received at
 
         """
-        values = self.stats.report()
-        self.stats_queue.put(values, True)
+        self.stats_queue.put(self.report_stats(), True)
         self.last_stats_time = time.time()
         signal.siginterrupt(signal.SIGPROF, False)
 
@@ -623,20 +638,35 @@ class Process(multiprocessing.Process, state.State):
             raise RuntimeError('Can not rejected messages when ack is False')
         if not self.can_respond:
             LOGGER.warning('Can not reject message, disconnected from RabbitMQ')
-            self.stats.incr(self.CLOSED_ON_COMPLETE)
-            if self.is_processing:
-                self.reset_state()
+            self.counters[self.CLOSED_ON_COMPLETE] += 1
+            self.measurement.set_tag(self.CLOSED_ON_COMPLETE, True)
             return
+
         LOGGER.warning('Rejecting message %s %s requeue', delivery_tag, 'with'
                        if requeue else 'without')
         self.channel.basic_nack(delivery_tag=delivery_tag, requeue=requeue)
-        if self.is_processing:
-            self.reset_state()
+        self.measurement.set_tag(self.NACKED, True)
+        self.measurement.set_tag(self.REQUEUED, requeue)
+
+    def report_stats(self):
+        """Create the dict of stats data for the MCP stats queue"""
+        if not self.previous:
+            self.previous = dict()
+            for key in self.counters:
+                self.previous[key] = 0
+        values = {
+            'name': self.name,
+            'consumer_name': self.consumer_name,
+            'counts': dict(self.counters),
+            'previous': dict(self.previous)
+        }
+        self.previous = dict(self.counters)
+        return values
 
     def reset_error_counter(self):
         """Reset the error counter to 0"""
         LOGGER.debug('Resetting the error counter')
-        self.stats[self.ERROR] = 0
+        self.counters[self.ERROR] = 0
 
     def reset_state(self):
         """Reset the runtime state after processing a message to either idle
@@ -644,12 +674,13 @@ class Process(multiprocessing.Process, state.State):
 
         """
         self.active_message = None
+        self.measurement = None
         if self.is_waiting_to_shutdown:
             self.set_state(self.STATE_SHUTTING_DOWN)
             self.on_ready_to_stop()
         elif self.is_processing:
             self.set_state(self.STATE_IDLE)
-        elif self.is_idle or self.is_connecting:
+        elif self.is_idle or self.is_connecting or self.is_shutting_down:
             pass
         else:
             LOGGER.critical('Unexepected state: %s', self.state_description)
@@ -677,12 +708,8 @@ class Process(multiprocessing.Process, state.State):
         except (AttributeError, ImportError) as error:
             name = self._kwargs['consumer_name']
             class_name = self._kwargs['config']['Consumers'][name]['consumer']
-            if PYTHON26:
-                LOGGER.exception('Could not start %s, stopping process: %r',
-                                 class_name, error)
-            else:
-                LOGGER.exception('Could not start %s, stopping process: %r',
-                                 class_name, error, exc_info=sys.exc_info())
+            LOGGER.exception('Could not start %s, stopping process: %r',
+                             class_name, error)
             os.kill(os.getppid(), signal.SIGABRT)
             sys.exit(1)
 
@@ -717,7 +744,7 @@ class Process(multiprocessing.Process, state.State):
                   'extra': {
                       'consumer_name': self.consumer_name,
                       'connection': self.connection_name,
-                      'env': self.strip_uri_passwords(dict(os.environ)),
+                      'env': dict(os.environ),
                       'message': message},
                   'time_spent': duration}
         LOGGER.debug('Sending exception to sentry: %r', kwargs)
@@ -736,20 +763,35 @@ class Process(multiprocessing.Process, state.State):
         LOGGER.info('Initializing for %s on %s connection', self.name,
                     connection_name)
 
-        # Setup the Sentry client
-        if raven and 'sentry_dsn' in cfg:
-            options = {
-                'include_paths': ['pika', 'rejected', 'tornado',
+        # Setup the Sentry client if configured and installed
+        sentry_dsn = cfg['Consumers'][consumer_name].get('sentry_dsn',
+                                                         cfg.get('sentry_dsn'))
+        if raven and sentry_dsn:
+            kwargs = {
+                'exclude_paths': ['tornado'],
+                'include_paths': ['pika',
+                                  'rejected',
                                   cfg['Consumers'][consumer_name]['consumer']],
+                'ignore_exceptions': ['rejected.consumer.ConsumerException',
+                                      'rejected.consumer.MessageException',
+                                      'rejected.consumer.ProcessingException'],
+                'processors': ['raven.processors.SanitizePasswordsProcessor']
             }
-            self.sentry_client = raven.Client(cfg['sentry_dsn'], **options)
+
+            if os.environ.get('ENVIRONMENT'):
+                kwargs['environment'] = os.environ['ENVIRONMENT']
+
+            if self.consumer_version:
+                kwargs['version'] = self.consumer_version
+
+            self.sentry_client = raven.Client(sentry_dsn, **kwargs)
 
         self.connection_name = connection_name
         self.consumer_name = consumer_name
         self.config = cfg['Consumers'][consumer_name]
+
         self.connections = cfg['Connections']
         self.consumer = self.get_consumer(self.config)
-        self.stats_queue = stats_queue
 
         if not self.consumer:
             LOGGER.critical('Could not import and start processor')
@@ -757,15 +799,9 @@ class Process(multiprocessing.Process, state.State):
             self.on_ready_to_stop()
             return
 
-        # Setup the stats counter instance
-        self.stats = stats.Stats(self.name, consumer_name, cfg['statsd'] or {})
+        self.stats_queue = stats_queue
 
-        # Set statsd in the consumer
-        if self.stats.statsd:
-            try:
-                self.consumer._set_statsd(self.stats.statsd)
-            except AttributeError:
-                LOGGER.info('Consumer does not support statsd assignment')
+        self.setup_instrumentation(cfg)
 
         # Consumer settings
         self.ack = self.config.get('ack', True)
@@ -794,12 +830,69 @@ class Process(multiprocessing.Process, state.State):
 
         # Setup QoS, Send a Basic.Recover and then Basic.Consume
         self.channel.basic_qos(self.on_qos_set, 0, self.qos_prefetch, False)
-        # self.channel.basic_recover(requeue=True)
         self.channel.basic_consume(consumer_callback=self.on_message,
                                    queue=self.queue_name,
                                    no_ack=not self.ack,
                                    consumer_tag=self.name)
 
+    def setup_influxdb(self, config):
+        """Configure the InfluxDB module for measurement submission.
+
+        :param dict config: The InfluxDB configuration stanza
+
+        """
+        base_tags = {
+            'connection': self.connection_name,
+            'version': self.consumer_version
+        }
+        measurement = self.config.get('influxdb_measurement',
+                                      os.environ.get('SERVICE'))
+        if measurement != self.consumer_name:
+            base_tags['consumer'] = self.consumer_name
+        for key in {'ENVIRONMENT', 'SERVICE'}:
+            if key in os.environ:
+                base_tags[key.lower()] = os.environ[key]
+        influxdb.install(
+            '{}://{}:{}/write'.format(
+                config.get('scheme',
+                           os.environ.get('INFLUXDB_SCHEME', 'http')),
+                config.get('host',
+                           os.environ.get('INFLUXDB_HOST', 'localhost')),
+                config.get('port', os.environ.get('INFLUXDB_PORT', '8086'))
+            ),
+            config.get('user', os.environ.get('INFLUXDB_USER')),
+            config.get('password', os.environ.get('INFLUXDB_PASSWORD')),
+            base_tags=base_tags)
+        return config.get('database', 'rejected'), measurement
+
+    def setup_instrumentation(self, config):
+        """Configure instrumentation for submission per message measurements
+        to statsd and/or InfluxDB.
+
+        :param dict config: The application configuration stanza
+
+        """
+        if not config.get('stats'):
+            if not config.get('statsd'):  # Backwards compatible statsd check
+                return
+            config['stats'] = {}
+
+        # Backwards compatible statsd config support
+        if config.get('statsd'):
+            warnings.warn('Deprecated statsd configuration detected',
+                          DeprecationWarning)
+            config['stats'].setdefault('statsd', config.get('statsd'))
+
+        if config['stats'].get('statsd'):
+            self.statsd = statsd.Client(self.consumer_name,
+                                        config['stats']['statsd'])
+            LOGGER.debug('statsd measurements configured')
+
+        # InfluxDB support
+        if influxdb and config['stats'].get('influxdb'):
+            self.influxdb = self.setup_influxdb(config['stats']['influxdb'])
+            LOGGER.debug('InfluxDB measurements configured: %r', self.influxdb)
+
     def setup_sighandlers(self):
         """Setup the stats and stop signal handlers."""
         signal.signal(signal.SIGINT, signal.SIG_IGN)
@@ -819,14 +912,6 @@ class Process(multiprocessing.Process, state.State):
         """
         self.message_connection_id = self.connection_id
 
-    @staticmethod
-    def strip_uri_passwords(values):
-        for key in values.keys():
-            matches = URI_RE.search(values[key])
-            if matches:
-                values[key] = values[key].replace(matches.group(1), '****')
-        return values
-
     def stop(self, signum=None, _unused=None):
         """Stop the consumer from consuming by calling BasicCancel and setting
         our state.
@@ -857,10 +942,6 @@ class Process(multiprocessing.Process, state.State):
                 signal.siginterrupt(signal.SIGTERM, False)
             return
 
-        # Stop and flush the statds data
-        if self.stats.statsd:
-            self.stats.statsd.stop()
-
         self.on_ready_to_stop()
 
     def stop_consumer(self):
@@ -874,6 +955,41 @@ class Process(multiprocessing.Process, state.State):
         except AttributeError:
             LOGGER.debug('Consumer does not have a shutdown method')
 
+    def submit_influxdb_measurement(self):
+        """Submit a measurement for a message to InfluxDB"""
+        measurement = influxdb.Measurement(*self.influxdb)
+        measurement.set_timestamp(time.time())
+        for key, value in self.measurement.counters.items():
+            measurement.set_field(key, value)
+        for key, value in self.measurement.tags.items():
+            measurement.set_tag(key, value)
+        for key, value in self.measurement.values.items():
+            measurement.set_field(key, value)
+        influxdb.add_measurement(measurement)
+        LOGGER.debug('InfluxDB Measurement: %r', measurement.marshall())
+
+    def submit_statsd_measurements(self):
+        """Submit a measurement for a message to statsd as individual items."""
+        for key, value in self.measurement.counters.items():
+            self.statsd.incr(key, value)
+        for key, value in self.measurement.values.items():
+            if isinstance(value, float):
+                self.statsd.add_timing(key, value)
+            else:
+                self.statsd.set_gauge(key, value)
+        for key, value in self.measurement.tags.items():
+            if isinstance(value, bool):
+                if value:
+                    self.statsd.incr(key)
+            elif isinstance(value, str):
+                if value:
+                    self.statsd.incr('{}.{}'.format(key, value))
+            elif isinstance(value, int):
+                self.statsd.incr(key, value)
+            else:
+                LOGGER.warning('The %s value type of %s is unsupported',
+                               key, type(value))
+
     @property
     def too_many_errors(self):
         """Return a bool if too many errors have occurred.
@@ -881,4 +997,4 @@ class Process(multiprocessing.Process, state.State):
         :rtype: bool
 
         """
-        return self.stats[self.ERROR] >= self.max_error_count
+        return self.counters[self.ERROR] >= self.max_error_count
